from __future__ import print_function
from Cython.Distutils import build_ext
import subprocess
import os
import sys


class Autogen(build_ext, object):
    def run(self, *args, **kwargs):
        self.autogen()
        super(Autogen, self).run(*args, **kwargs)

    def autogen(self):
        NAME = 'admesh'
        HEADERS = [
            'admesh/stl.h',
        ]
        PREFIX = 'stl_'
        PRECUT = len(PREFIX)
        PXD_IGNORE = [
            '',
            'exit_on_error',
            'stats_out',
            'print_neighbors',
            'print_edges',
            'scale_versor',
            'add_facet',
            'allocate',
            'reallocate',
            'count_facets',
            'write_binary_block',
            'put_little_int',
            'put_little_float',
        ]
        PYX_IGNORE = ['open', 'close', 'get_error', 'clear_error', 'repair']

        PXD = 'c' + NAME + '.pxd'
        _PXD = '_' + PXD
        PYX = NAME + '.pyx'
        _PYX = '_' + PYX

        SELF = 'stl_file *stl'

        print('generating {PXD} and {PYX}'.format(PXD=PXD, PYX=PYX))

        pxdlines = []
        pyxlines = []

        for header in HEADERS:
            _header = self.get_header(header)
            if not _header:
                sys.stderr.write('Error: {h} not found, install ADMesh first.\n\nSee README.rst'
                                 'for more information\n'.format(h=header))
                exit(1)
            with open(_header) as h:
                lines = h.readlines()
            for line in lines:
                if not line:
                    continue
                line = line[:-2] # cut off ;
                line = line.split()
                if not line or line[0] != 'extern':
                    continue
                pxd_line = '    ' + ' '.join(line[1:]) + '\n' # void stl_foo(...)
                line = ' '.join(line[2:]) # stl_foo(...)
                line = line.split('(')
                function = line[0][PRECUT:]
                if function in PXD_IGNORE:
                    continue
                pxdlines.append(pxd_line)
                if function in PYX_IGNORE:
                    continue
                line = line[1][:-1]
                line = [l.strip() for l in line.split(',')]
                static = True
                if line[0] == SELF:
                    static = False
                    line = line[1:]
                if static:
                    continue # this is a hack, we don' want any static method NOW
                args = []
                chars = set()
                for arg in line:
                    char = arg.split()[-2]
                    arg = arg.split()[-1]
                    if arg.startswith('*'):
                        arg = arg[1:]
                    if arg.endswith('[]'):
                        arg = arg[:-2]
                    if arg.endswith('[3]'):
                        arg = arg[:-3]
                    args.append(arg)
                    if char == 'char':
                        chars.add(arg)
                if static:
                    pyxlines.append('    @classmethod\n')
                pyxlines.append('    def ')
                pyxlines.append(function)
                pyxlines.append('(')
                if static:
                    pyxlines.append('cls')
                else:
                    pyxlines.append('self')
                if args:
                    pyxlines.append(', ')
                    label = -1
                    for i, arg in enumerate(args):
                        if arg == 'label':
                            label = i
                    if label >= 0:
                        largs = args[:label] + args[label+1:] + ["label='admesh'"]
                    else:
                        largs = args
                    pyxlines.append(', '.join(largs))
                pyxlines.append('):')
                pyxlines.append('\n        """')
                pyxlines.append(PREFIX+function)
                pyxlines.append('"""')
                pyxlines.append('\n        if not self._opened:')
                pyxlines.append('\n            raise AdmeshError(\'STL not opened\')')
                for idx, arg in enumerate(args):
                    if arg not in chars:
                        continue
                    pyxlines.append('\n        ')
                    pyxlines.append('by_{arg} = {arg}.encode(\'UTF-8\')'.format(arg=arg))
                    args[idx] = 'by_' + arg
                pyxlines.append('\n        ')
                pyxlines.append(PREFIX+function)
                pyxlines.append('(')
                if not static:
                    pyxlines.append('&self._c_stl_file')
                    if args:
                        pyxlines.append(', ')
                if args:
                    pyxlines.append(', '.join(args))
                pyxlines.append(')\n')
                pyxlines.append('        if stl_get_error(&self._c_stl_file):\n')
                pyxlines.append('            stl_clear_error(&self._c_stl_file)\n')
                pyxlines.append('            raise AdmeshError(\'')
                pyxlines.append(PREFIX+function)
                pyxlines.append('\')\n')
                pyxlines.append('\n')
                

        with open(_PXD, 'r') as _pxd:
            _pxdlines = _pxd.readlines() + ['\n']

        with open(PXD, 'w') as pxd:
            pxd.write(''.join(_pxdlines + pxdlines))
            pxd.write('\n')

        with open(_PYX, 'r') as _pyx:
            _pyxlines = _pyx.readlines() + ['\n']

        with open(PYX, 'w') as pyx:
            pyx.write(''.join(_pyxlines + pyxlines))
            pyx.write('\n')

    def get_header(self, header):
        cflags = os.environ.get('CFLAGS','')
        p = subprocess.Popen('gcc -v -E -'.split() + cflags.split(), stdin=subprocess.PIPE, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        out, log = p.communicate('')
        log = log.decode('utf8').split('\n')
        read = False
        for line in log:
            if 'search starts here:' in line:
                read = True
                continue
            if 'End of search list' in line:
                break
            if read:
                candidate = os.path.join(line.strip(),header)
                if os.path.isfile(candidate):
                    print("found %s" % candidate)
                    return candidate
        return None




from cadmesh cimport *


class AdmeshError(Exception):
    pass


cdef class Stl:
    cdef stl_file _c_stl_file
    cdef bint _opened
    def __cinit__(self, path=''):
        self._opened = False
        if path:
            self.open(path)

    def open(self, path):
        """stl_open"""
        by_path = path.encode('UTF-8')
        stl_open(&self._c_stl_file, by_path)
        if stl_get_error(&self._c_stl_file):
            stl_clear_error(&self._c_stl_file)
            raise AdmeshError('stl_open')
        self._opened = True

    def repair(self,
               fixall_flag=True,
               exact_flag=False,
               tolerance_flag=False,
               tolerance=0,
               increment_flag=False,
               increment=0,
               nearby_flag=False,
               iterations=2,
               remove_unconnected_flag=False,
               fill_holes_flag=False,
               normal_directions_flag=False,
               normal_values_flag=False,
               reverse_all_flag=False,
               verbose_flag=True):
        """stl_repair"""
        if not self._opened:
            raise AdmeshError('STL not opened')
        stl_repair(&self._c_stl_file,
                   fixall_flag,
                   exact_flag,
                   tolerance_flag,
                   tolerance,
                   increment_flag,
                   increment,
                   nearby_flag,
                   iterations,
                   remove_unconnected_flag,
                   fill_holes_flag,
                   normal_directions_flag,
                   normal_values_flag,
                   reverse_all_flag,
                   verbose_flag)
        if stl_get_error(&self._c_stl_file):
            stl_clear_error(&self._c_stl_file)
            raise AdmeshError('stl_repair')


    def __dealloc__(self):
        if self._opened:
            stl_close(&self._c_stl_file)



# -*- coding: utf-8 -*-
from admesh import Stl
import filecmp


class TestIO(object):
    '''Tests for the  basic IO operations'''
    def test_saved_equals_original_ascii(self):
        '''Tests if saved ASCII file is identical to the loaded one'''
        stl = Stl('test/block.stl')
        stl.write_ascii('test/block_ascii.stl')
        assert filecmp.cmp('test/block.stl', 'test/block_ascii.stl')

    def test_saved_equals_original_binary(self):
        '''Tests if saved binary file is identical to the loaded one'''
        stl1 = Stl('test/block.stl')
        stl1.write_binary('test/block_binary.stl')
        stl2 = Stl('test/block_binary.stl')
        stl2.write_binary('test/block_binary2.stl')
        assert filecmp.cmp('test/block_binary.stl', 'test/block_binary2.stl')

    def test_save_load_unicode(self):
        '''Tests saving and loading files with Unicode filenames'''
        stl1 = Stl('test/block.stl')
        stl1.write_ascii(u'test/block_ěščřž.stl')
        stl2 = Stl(u'test/block_ěščřž.stl')



cdef extern from "admesh/stl.h":
    ctypedef struct stl_vertex:
        pass

    ctypedef struct stl_normal:
        pass

    ctypedef struct stl_facet:
        pass

    cdef enum stl_type:
        pass

    ctypedef struct stl_edge:
        pass

    ctypedef struct stl_hash_edge:
        pass

    ctypedef struct stl_neighbors:
        pass

    ctypedef struct v_indices_struct:
        pass

    ctypedef struct stl_stats:
        pass

    ctypedef struct stl_file:
        pass



#!/usr/bin/env python
# -*- coding: utf-8 -*-
from setuptools import setup, find_packages, Extension
from autogen import Autogen

long_description = ''.join(open('README.rst').readlines())

setup(
    name='admesh',
    version='0.98.2',
    description='Python bindings for ADMesh, STL maipulation library',
    long_description=long_description,
    keywords='STL, mesh, 3D',
    author='Miro Hrončok',
    author_email='miro@hroncok.cz',
    url='https://github.com/admesh/python-admesh',
    license='GPLv2+',
    install_requires=['Cython'],
    packages=find_packages(),
    cmdclass={'build_ext': Autogen},
    ext_modules=[Extension("admesh", ["admesh.pyx"], libraries=["admesh"])],
    classifiers=['Development Status :: 5 - Production/Stable',
                 'License :: OSI Approved :: GNU General Public License v2 or later (GPLv2+)',
                 'Operating System :: POSIX :: Linux',
                 'Operating System :: MacOS :: MacOS X',
                 'Programming Language :: Cython',
                 'Programming Language :: Python :: 2',
                 'Programming Language :: Python :: 2.6',
                 'Programming Language :: Python :: 2.7',
                 'Programming Language :: Python :: 3',
                 'Programming Language :: Python :: 3.3',
                 'Programming Language :: Python :: 3.4',
                 'Programming Language :: Python :: Implementation :: CPython',
                 'Programming Language :: Python :: Implementation :: PyPy',
                 'Topic :: Multimedia :: Graphics :: 3D Modeling',
                ],
)
